SystemOrganization addCategory: #'MC2-Repository-File'!

MDRepository subclass: #MDFileRepository
	instanceVariableNames: 'dataStream index memos lastPosition'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repository-File'!

!MDFileRepository class methodsFor: 'accessing' stamp: 'cwp 10/12/2005 21:30'!
extension
	^ '.mcr'! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'jf 11/8/2009 19:15'!
id: aHashstamp name: aString
	| theName |
	theName := (aString endsWith: self extension)
		ifTrue: [aString]
		ifFalse: [aString, self extension].
	^ (self on: (FileStream fileNamed: theName))
		setId: aHashstamp;
		yourself! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'cwp 2/2/2006 22:20'!
named: aString
	^ self on: (FileStream fileNamed: aString, self extension)! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'dc 7/19/2007 10:41'!
named: aString do: aBlock
	^ self
		on: (FileStream fileNamed: aString, self extension)
		do: aBlock! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'jf 11/8/2009 19:13'!
on: aFileStream
	^ self basicNew
		initializeWithStream: aFileStream;
		yourself! !

!MDFileRepository class methodsFor: 'instance creation' stamp: 'cwp 10/22/2009 00:07'!
on: aFileStream do: aBlock 
	| repository |
	repository := self on: aFileStream.
	^ [ aBlock value: repository ] 
		ensure: [ repository close ]! !

!MDFileRepository class methodsFor: 'class initialization' stamp: 'cwp 10/22/2008 20:47'!
startUp
	"Do nothing. This is implemented because old versions registered
	to be notified at start up, and that registration might still be in effect."! !

!MDFileRepository methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:22'!
= other
	^ self species = other species
		and: [other dataStreamIs: dataStream]! !

!MDFileRepository methodsFor: 'synchronization' stamp: 'cwp 12/23/2007 00:11'!
allHashstamps
	self rescanData.
	^ index keys collect: [:ea | MDHashstamp fromBytes: ea]! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 10/13/2005 23:01'!
close
	dataStream close! !

!MDFileRepository methodsFor: 'private' stamp: 'cwp 12/25/2006 15:23'!
dataStreamIs: aStream
	^ dataStream = aStream! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 10/22/2009 12:18'!
destroy
	| path |
	super destroy.
	path := dataStream stream fullName.
	FileDirectory deleteFilePath: path.! !

!MDFileRepository methodsFor: 'converting' stamp: 'cwp 10/17/2009 23:22'!
elementClass
	^ MDFileRepositoryElement! !

!MDFileRepository methodsFor: 'private' stamp: 'cwp 10/22/2008 20:44'!
flushIndex	
	lastPosition := 0.
	index := Dictionary new.
	memos := Set new.
	! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 11/8/2009 21:43'!
get: aHashstamp 
	| key bytes position |
	key := aHashstamp bytes.
	position := index at: key ifAbsent: [^self rescanData ifTrue: [self get: aHashstamp]].
	bytes := dataStream valueAtPosition: position.
	bytes ifNil:
		[self flushIndex.
		self rescanData.
		bytes := dataStream valueAtPosition: position].
	^ bytes ifNotNil: [MDMaterializer materialize: bytes]! !

!MDFileRepository methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:25'!
hash
	^ dataStream hash! !

!MDFileRepository methodsFor: 'synchronization' stamp: 'cwp 12/23/2007 00:12'!
includes: aHashstamp
	self rescanData.
	^ index includesKey: aHashstamp bytes! !

!MDFileRepository methodsFor: 'initialize-release' stamp: 'cwp 11/8/2009 23:32'!
initializeWithStream: aFileStream
	self initialize. 
	self stream: aFileStream! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 1/19/2007 20:57'!
memos
	
	self rescanData.
	^ memos! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 5/31/2008 14:43'!
name
	^ dataStream stream localName! !

!MDFileRepository methodsFor: 'public' stamp: 'cwp 11/8/2009 21:46'!
put: aHashedObject
	| key value |
	key := aHashedObject hashstamp bytes.
	index at: key ifAbsent:
		[value := MDSerializer serialize: aHashedObject.
		index at: key put: dataStream nextValuePosition.
		dataStream appendKey: key value: value.
		dataStream flush]! !

!MDFileRepository methodsFor: 'private' stamp: 'cwp 11/8/2009 21:43'!
rescanData
	dataStream reopen.
	dataStream fromPosition: lastPosition keysValuesAndPositionsDo:  
		[:key :val :pos |
		index at: key put: pos.
		(MDMaterializer isMemo: val) ifTrue:
			[memos add: (MDMaterializer materialize: val)]].
	dataStream position = lastPosition
		ifTrue: [^ false]
		ifFalse: [lastPosition := dataStream position. ^ true]! !

!MDFileRepository methodsFor: 'initialize-release' stamp: 'cwp 11/8/2009 23:32'!
stream: aFileStream
	dataStream := MDKeyedStream on: aFileStream.
	lastPosition := 0.
	index := Dictionary new.
	memos := Set new.
	self rescanData! !

!MDFileRepository methodsFor: 'accessing' stamp: 'cwp 11/8/2009 23:32'!
updateName: newName
	| oldName |
	dataStream
		ifNil: [self stream: (FileDirectory default fileNamed: newName)]
		ifNotNil:
			[oldName := dataStream stream localName.
			oldName = newName ifFalse: 
				[ | directory |
				directory := dataStream stream directory.
				self close.
				directory rename: oldName toBe: newName.
				self stream: (directory oldFileNamed: newName).
				self setId: id]]! !

MDObject subclass: #MDByteIndex
	instanceVariableNames: 'keySize valueSize streamSize stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repository-File'!

!MDByteIndex class methodsFor: 'as yet unclassified' stamp: 'avi 3/15/2005 03:08'!
stream: aStream keySize: keyNumber valueSize: valueNumber
	^ self stream: aStream keySize: keyNumber valueSize: valueNumber streamSize: aStream size! !

!MDByteIndex class methodsFor: 'as yet unclassified' stamp: 'avi 3/19/2005 02:41'!
stream: aStream keySize: keyNumber valueSize: valueNumber streamSize: streamNumber
	^ self basicNew setStream: aStream keySize: keyNumber valueSize: valueNumber streamSize: streamNumber! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:36'!
at: keyArray
	^ self at: keyArray ifAbsent: [self error: 'Could not find key ', keyArray printString]! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:36'!
at: keyArray ifAbsent: errorBlock
	keyArray size = keySize ifFalse: [self error: 'Key must be exactly ', keySize asString, ' bytes'].
	^ self basicAt: keyArray ifAbsent: errorBlock! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:37'!
at: keyArray put: valueArray
	keyArray size = keySize ifFalse: [self error: 'Key must be exactly ', keySize asString, ' bytes'].
	valueArray size = valueSize ifFalse: [self error: 'Value must be exactly ', valueSize asString, ' bytes'].
	^ self basicAt: keyArray put: valueArray! !

!MDByteIndex methodsFor: 'private' stamp: 'cwp 5/22/2008 00:11'!
basicAt: keyArray ifAbsent: errorBlock
	| i key |
	i := self scanFor: keyArray.
	key := self keyBytesAtIndex: i.
	^ key = keyArray
		ifTrue: [self valueBytesAtIndex: i]
		ifFalse: errorBlock! !

!MDByteIndex methodsFor: 'private' stamp: 'cwp 5/22/2008 00:11'!
basicAt: keyArray put: valueArray
	| i key |
	i := self scanFor: keyArray.
	key := self keyBytesAtIndex: i.
	key = keyArray ifFalse:
		[self writeTally: self tally + 1.
		self writeKeyBytes: keyArray atIndex: i].
	self writeValueBytes: valueArray atIndex: i.
	^ valueArray! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:45'!
hash: keyArray
	^ keyArray hash! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/15/2005 03:38'!
keyBytesAtIndex: i
	self seekTo: i.
	^ stream atEnd
		ifTrue: [ByteArray new: keySize]
		ifFalse: [stream next: keySize]! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:05'!
keySize
	^ keySize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/15/2005 03:07'!
maxIndex
	^ streamSize - 4 // self recordSize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:43'!
recordSize
	^ keySize + valueSize! !

!MDByteIndex methodsFor: 'private' stamp: 'cwp 5/22/2008 00:11'!
scanFor: keyArray
	| start |
	start := (self hash: keyArray) \\ self maxIndex.
	start to: self maxIndex do: [:i | (self scanFor: keyArray at: i) ifTrue: [^ i]].
	0 to: start - 1 do: [:i | (self scanFor: keyArray at: i) ifTrue: [^ i]].
	
	self error: 'Index is full'! !

!MDByteIndex methodsFor: 'private' stamp: 'cwp 5/22/2008 00:11'!
scanFor: keyArray at: i
	| key |
	key := self keyBytesAtIndex: i.
	^ key = keyArray or: [key allSatisfy: [:b | b = 0]]! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:42'!
seekTo: i
	stream position: i * self recordSize + 4! !

!MDByteIndex methodsFor: 'initialize-release' stamp: 'cwp 5/22/2008 00:11'!
setStream: aStream keySize: keyNumber valueSize: valueNumber streamSize: streamNumber
	keySize := keyNumber.
	valueSize := valueNumber.
	streamSize := streamNumber.
	stream := aStream.
	stream binary! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/19/2005 02:51'!
streamSize
	^ streamSize! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/15/2005 03:23'!
tally
	stream reset.
	^ stream atEnd
		ifTrue: [0]
		ifFalse: [stream nextInt32]! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:41'!
valueBytesAtIndex: i
	self seekTo: i.
	stream skip: keySize.
	^ stream next: valueSize! !

!MDByteIndex methodsFor: 'accessing' stamp: 'avi 3/13/2005 03:05'!
valueSize
	^ valueSize! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:40'!
writeKeyBytes: keyBytes atIndex: i
	self seekTo: i.
	stream nextPutAll: keyBytes! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:42'!
writeTally: aNumber
	stream reset.
	stream nextInt32Put: aNumber! !

!MDByteIndex methodsFor: 'private' stamp: 'avi 3/13/2005 03:40'!
writeValueBytes: valueBytes atIndex: i
	self seekTo: i.
	stream skip: keySize.
	stream nextPutAll: valueBytes! !

MDObject subclass: #MDKeyedStream
	instanceVariableNames: 'stream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MC2-Repository-File'!

!MDKeyedStream class methodsFor: 'as yet unclassified' stamp: 'avi 4/17/2005 01:19'!
on: aStream
	^ self basicNew initializeWithStream: aStream! !

!MDKeyedStream methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:25'!
= other
	^ self species = other species
		and: [self stream name = other stream name]! !

!MDKeyedStream methodsFor: 'public' stamp: 'cwp 11/6/2007 22:52'!
appendKey: keyBytes value: valueBytes 
	self assert: [keyBytes size = self keySize].
	stream
		setToEnd;
		nextPutAll: keyBytes;
		nextInt32Put: valueBytes size;
		nextPutAll: valueBytes! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 10/13/2005 23:00'!
close
	stream name = 'a stream' ifFalse: [stream close]! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 01:58'!
flush
	stream flush! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/19/2005 01:49'!
fromPosition: aNumber keysValuesAndPositionsDo: aBlock 
	| position key |
	stream position: aNumber.
	[stream atEnd] whileFalse: 
			[key := stream next: self keySize.
			position := stream position.
			aBlock
				value: key
				value: self readValue
				value: position]! !

!MDKeyedStream methodsFor: 'comparing' stamp: 'cwp 12/25/2006 15:25'!
hash
	^ self stream name hash! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 5/22/2008 00:11'!
initializeWithStream: aPositionableStream
	stream := aPositionableStream.
	stream binary! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 4/10/2005 15:14'!
keySize
	^ 24! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 02:11'!
keysValuesAndPositionsDo: aBlock 
	self fromPosition: 0 keysValuesAndPositionsDo: aBlock
			! !

!MDKeyedStream methodsFor: 'public' stamp: 'avi 4/17/2005 01:38'!
nextValuePosition
	^ stream size + self keySize! !

!MDKeyedStream methodsFor: 'public' stamp: 'cwp 4/10/2005 17:29'!
position
	^ stream position! !

!MDKeyedStream methodsFor: 'private' stamp: 'cwp 10/22/2008 20:44'!
readValue
	| length |
	[length := stream nextInt32]
		on: Error
		do: [:error | ^ nil].
	^ stream next: length.
	! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'avi 3/22/2006 19:23'!
reopen
	stream name = 'a stream' ifFalse: [stream reopen; binary]! !

!MDKeyedStream methodsFor: 'accessing' stamp: 'cwp 4/10/2005 14:53'!
stream
	^ stream! !

!MDKeyedStream methodsFor: 'public' stamp: 'cwp 10/31/2008 23:49'!
valueAtPosition: anInteger
	[stream position: anInteger]
		on: Error
		do: [:err | ^ nil].
	^ self readValue! !
